library lwws;

uses Windows, WinSock, NLib, NSys, NUtil, NSock, NkMonAPI; 

{$E nmp}

type
  TEvents = class(TObject)
    procedure OnGetHostByName(const IpAddr: string);
    procedure OnConnect(Sender: TObject);
    procedure OnRecv(Sender: TObject);
    procedure OnClose(Sender: TObject);
    procedure OnError(Sender: TObject; ErrorCode: Integer);
  end;

var
  Tcp: TNTcp;
  Events: TEvents;
  Buf: string;
  FDay: string;
  Server: string;
  Path: string;
  City: string;

function UTF8toSJIS(const S: string): string;
var
  w: string;
  count: Integer;
begin
  SetString(w, nil, Length(S)*2);
  count := MultiByteToWideChar(CP_UTF8, 0, PChar(S), Length(S), PWideChar(w),
    Length(w) div 2); 
  SetLength(w, count*2);
  SetString(Result, nil, count*2);
  SetLength(Result, WideCharToMultiByte(CP_ACP, 0, PWideChar(w), count,
    PChar(Result), count*2, nil, nil));
end;

procedure SetValue(const Name: string; Value: string);
begin
  if Value = '' then
    Value := '--';
  nkmSetValue(HInstance, PChar(FDay+'.'+Name), PChar(Value));
end;

function GetAttr(const S, Name: string): string;
var
  i: Integer;
  q: Boolean;
  attr, token: string;
begin
  Result := '';
  i := 0;
  q := False;
  while i < Length(S) do
  begin
    Inc(i);
    case S[i] of
      '"': q := not q;
      ' ':
        if q then
          token := token + S[i]
        else if attr = Name then
        begin
          Result := token;
          Exit;
        end else
          token := '';
      '=':
        begin
          attr := token;
          token := '';
        end;
    else
       token := token + S[i];
    end;
  end;
end;

procedure Parse(const S: string);
var
  i, j: Integer;
  intag, skip: Boolean;
  tag, attr, cont, name: string;
  tags: TStrList;
begin
  i := 0;
  intag := False;
  skip := False;
  tags := TStrList.Create;
  while i < Length(S) do
  begin
    Inc(i);
    case S[i] of
      '<':
        begin
          intag := True;
          skip := False;
        end;
      '>':
        begin
          intag := False;
          if (tag = '?xml') or ((i > 1) and (S[i-1] = '/')) then
          begin
            if tag = 'location' then
              SetValue('city', Utf8ToSjis(GetAttr(attr, 'city')));
            attr := '';
            tag := '';
            continue;
          end;
          if tag[1] = '/' then
          begin
            name := tags[0];
            for j := 1 to tags.Count-1 do
              name := name + '.' + tags[j];
            cont := Utf8ToSjis(cont);
            if name = 'lwws.day' then
              SetValue('day', Copy(cont, 1, 3))
            else if name = 'lwws.publictime' then
            begin
              SetValue('publictime.day', Copy(cont, 6, 2));
              SetValue('publictime.hour', Copy(cont, 18, 2));
              SetValue('publictime.min', Copy(cont, 21, 2));
            end
            else if name = 'lwws.telop' then
              SetValue('telop', cont)
            else if name = 'lwws.temperature.max.celsius' then
              SetValue('maxtemp', cont)
            else if name = 'lwws.temperature.min.celsius' then
              SetValue('mintemp', cont);
            cont := '';
            tags.Delete(tags.Count-1);
          end else
            tags.Add(tag);
          tag := '';
        end;
      #10: ; // pass 
    else
      if intag then
      begin
        if S[i] = ' ' then
          skip := True;
        if skip then
          attr := attr + S[i]
        else
          tag := tag + S[i]
      end else
        cont := cont + S[i];
    end;
  end;
  tags.Free;
end;

procedure TEvents.OnGetHostByName(const IpAddr: string);
begin
  if IpAddr = '' then
    nkmLog(HInstance, PChar(Server+': Host not found'))
  else begin
    nkmLog(HInstance, PChar('Connecting to '+Server+'...'));
    Tcp.Connect(IpAddr, 80);
  end;
end;

procedure TEvents.OnConnect(Sender: TObject);
begin
  nkmLog(HInstance, 'Connected. Sending HTTP request...');
  (Sender as TNTcp).Send('GET '+Path+
      '?city='+City+'&day='+FDay+' HTTP/1.0'#13#10#13#10);
end;

procedure TEvents.OnRecv(Sender: TObject);
var
  s: string;
begin
  repeat
    s := (Sender as TNTcp).Recv(1024);
    Buf := Buf + s;
  until s = '';
end;

procedure TEvents.OnError(Sender: TObject; ErrorCode: Integer);
begin
  nkmLog(HInstance, PChar(SysError(ErrorCode)));
end;

procedure TEvents.OnClose(Sender: TObject);
begin
  nkmLog(HInstance, 'Received response.');
  Parse(Buf);
  Buf := '';
  if FDay = 'today' then
  begin
    FDay := 'tomorrow';
    Tcp.GetHostByName(Server);
  end
  else if FDay = 'tomorrow' then
  begin
    FDay := 'dayaftertomorrow';
    Tcp.GetHostByName(Server);
  end;
end;

function gauge_init: Integer; stdcall;
begin
  Server := nkmConf(HInstance, 'server', 'weather.livedoor.com');
  Path := nkmConf(HInstance, 'path', '/forecast/webservice/rest/v1');
  City := nkmConf(HInstance, 'city', '63'); // default is Tokyo
  Events := TEvents.Create;
  Tcp := TNTcp.Create;
  Tcp.OnGetHostByName := Events.OnGetHostByName;
  Tcp.OnConnect := Events.OnConnect;
  Tcp.OnRecv := Events.OnRecv;
  Tcp.OnClose := Events.OnClose;
  Tcp.OnError := Events.OnError;
  Result := 3600;
end;

procedure gauge_update; stdcall;
begin
  FDay := 'today';
  Tcp.GetHostByName(Server);
end;

procedure gauge_unload; stdcall;
begin
  Tcp.Free;
  Events.Free;
end;

exports
  gauge_init,
  gauge_update,
  gauge_unload;

end.
