unit SmartHelper;

interface

uses Windows, Smart, NLib;

type
  PDriveAttributeArray = ^TDriveAttributeArray;
  TDriveAttributeArray =
    array[0..MaxInt div SizeOf(TDriveAttribute)-1] of TDriveAttribute;
  TSmart = class(TObject)
  private
    FDrv: Integer;
    AttrBuffer: PChar;
    IdBuffer: PChar;
  public
    Handle: THandle;
    Id: PIdSector;
    Attr: PDriveAttributeArray;
    constructor Create(Drv: Integer);
    destructor Destroy; override;
    function ReadAttrValues: Boolean;
    function ReadIdSector: Boolean;
  end;

implementation

procedure SwapByteOrder(buf: PChar; len: Integer);
var
  i: Integer;
  c: Char;
begin
  i := 0;
  while i < len do
  begin
    c := buf[i];
    buf[i] := buf[i+1];
    buf[i+1] := c;
    Inc(i, 2);
  end;
end;

constructor TSmart.Create(Drv: Integer);
var
  path: string;
begin
  FDrv := Drv;
  path := '\\.\PhysicalDrive'+IStr(FDrv);
  // -1 for TSendCmdOutParams.Buffer[0]
  GetMem(AttrBuffer, SizeOf(TSendCmdOutParams)+READ_ATTRIBUTE_BUFFER_SIZE-1);
  GetMem(IdBuffer, SizeOf(TSendCmdOutParams)+IDENTIFY_BUFFER_SIZE-1);
  Handle := CreateFile(PChar(path),
    GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE,
    nil, OPEN_EXISTING, 0, 0);
  if Handle = INVALID_HANDLE_VALUE then
    raise EIOError.Create(path);
end;

destructor TSmart.Destroy;
begin
  if Handle <> INVALID_HANDLE_VALUE then
    CloseHandle(Handle);
  FreeMem(AttrBuffer);
  FreeMem(IdBuffer);
end;

function TSmart.ReadAttrValues: Boolean;
var
  p: TSendCmdInParams;
  retsize: DWORD;
begin
  p.cBufferSize := READ_ATTRIBUTE_BUFFER_SIZE;
  p.irDriveRegs.bFeaturesReg := SMART_READ_ATTRIBUTE_VALUES;
  p.irDriveRegs.bSectorCountReg := 1;
  p.irDriveRegs.bSectorNumberReg := 1;
  p.irDriveRegs.bCylLowReg := SMART_CYL_LOW;
  p.irDriveRegs.bCylHighReg := SMART_CYL_HI;
  p.irDriveRegs.bDriveHeadReg := $A0 or ((FDrv and 1) shl 4);
  p.irDriveRegs.bCommandReg := IDE_EXECUTE_SMART_FUNCTION;
  p.bDriveNumber := FDrv;
  // -1 for TSendCmdOutParams.Buffer[0]
  Result := DeviceIoControl(Handle, DFP_RECEIVE_DRIVE_DATA, @p, SizeOf(p)-1,
    AttrBuffer, SizeOf(TSendCmdOutParams)+READ_ATTRIBUTE_BUFFER_SIZE-1,
    retsize, nil);
  // avoid bBuffer range check
  Attr := @(PChar(@PSendCmdOutParams(AttrBuffer).bBuffer[0])[2]);
end;

function TSmart.ReadIdSector;
var
  p: TSendCmdInParams;
  retsize: DWORD;
begin
  p.cBufferSize := IDENTIFY_BUFFER_SIZE;
  p.irDriveRegs.bFeaturesReg := 0;
  p.irDriveRegs.bSectorCountReg := 1;
  p.irDriveRegs.bSectorNumberReg := 1;
  p.irDriveRegs.bCylLowReg := 0;
  p.irDriveRegs.bCylHighReg := 0;
  p.irDriveRegs.bDriveHeadReg := $A0 or ((FDrv and 1) shl 4);
  p.irDriveRegs.bCommandReg := IDE_ID_FUNCTION; //IDE_ATAPI_ID;
  p.bDriveNumber := FDrv;
  // -1 for TSendCmdOutParams.Buffer[0]
  Result := DeviceIoControl(Handle, DFP_RECEIVE_DRIVE_DATA, @p, SizeOf(p)-1,
    IdBuffer, SizeOf(TSendCmdOutParams)+IDENTIFY_BUFFER_SIZE-1, retsize, nil);
  Id := @PSendCmdOutParams(IdBuffer).bBuffer[0];
  SwapByteOrder(Id.sModelNumber, SizeOf(Id.sModelNumber));
end;

end.
